/*
 * 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 2007-2008 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 *
 */

 

package com.sun.grizzly.util.http;

import java.io.Serializable;
import java.text.FieldPosition;
import java.util.Date;

import com.sun.grizzly.util.buf.ByteChunk;
import com.sun.grizzly.util.buf.DateTool;
import com.sun.grizzly.util.buf.MessageBytes;
import com.sun.grizzly.util.res.StringManager;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 *  Server-side cookie representation.
 *  Allows recycling and uses MessageBytes as low-level
 *  representation ( and thus the byte-> char conversion can be delayed
 *  until we know the charset ).
 *
 *  Tomcat.core uses this recyclable object to represent cookies,
 *  and the facade will convert it to the external representation.
 */
public class ServerCookie implements Serializable {

    private static Logger logger = Logger.getLogger("Grizzly");
    private static final StringManager sm = StringManager.getManager(
        "com.sun.grizzly.util.http.res");

    // START PWC 6392327
    /**
     * The string sent as the value of the cookie in the Set-Cookie header
     * for a Cookie whose value is null.
     */
    private static final String NULL_VALUE = "null";
    // END PWC 6392327

    private static final String ANCIENT_DATE =
        DateTool.formatOldCookie(new Date(10000));

    // Version 0 (Netscape) attributes
    private MessageBytes name=MessageBytes.newInstance();
    private MessageBytes value=MessageBytes.newInstance();
    // Expires - Not stored explicitly. Generated from Max-Age (see V1)
    private MessageBytes path = MessageBytes.newInstance();	// ;Path=VALUE .
    private MessageBytes domain = MessageBytes.newInstance();    // ;Domain=VALUE ...
    private boolean secure;	// ;Secure

    // Version 1 (RFC2109) attributes
    private MessageBytes comment=MessageBytes.newInstance();    // ;Comment=VALUE
    private int maxAge = -1;	// ;Max-Age=VALUE
				// ;Discard ... implied by maxAge < 0
    // RFC2109: maxAge=0 will end a session

    private int version = 0;	// ;Version=1

    // Note: Servlet Spec =< 2.5 only refers to Netscape and RFC2109,
    // not RFC2965

    // Version 1 (RFC2965) attributes
    // TODO Add support for CommentURL
    // Discard - implied by maxAge <0
    // TODO Add support for Port

    public ServerCookie() {
    }

    public void recycle() {
        path.recycle();
        name.recycle();
        value.recycle();
        comment.recycle();
        maxAge=-1;
        path.recycle();
        domain.recycle();
        version=0;
        secure=false;
    }

    public MessageBytes getComment() {
	return comment;
    }

    public MessageBytes getDomain() {
	return domain;
    }

    public void setMaxAge(int expiry) {
	maxAge = expiry;
    }

    public int getMaxAge() {
	return maxAge;
    }


    public MessageBytes getPath() {
	return path;
    }

    public void setSecure(boolean flag) {
	secure = flag;
    }

    public boolean getSecure() {
	return secure;
    }

    public MessageBytes getName() {
	return name;
    }

    public MessageBytes getValue() {
	return value;
    }

    public int getVersion() {
	return version;
    }


    public void setVersion(int v) {
	version = v;
    }


    // -------------------- utils --------------------

    public String toString() {
	return "Cookie " + getName() + "=" + getValue() + " ; "
	    + getVersion() + " " + getPath() + " " + getDomain();
    }
    
    private static final String tspecials = ",; ";
    private static final String tspecials2 = "()<>@,;:\\\"/[]?={} \t";
    // work around for IE not being RFC complaint
    // IE doesn't like quotes around the path value, even though a slash
    // is considered a header delimiter and must be quoted
    private static final String tspecials2NoSlash = "()<>@,;:\\\"[]?={} \t";

    /*
     * Tests a string and returns true if the string counts as a
     * reserved token in the Java language.
     *
     * @param value		the <code>String</code> to be tested
     *
     * @return			<tt>true</tt> if the <code>String</code> is
     *				a reserved token; <tt>false</tt>
     *				if it is not
     */
    public static boolean isToken(String value) {
        return isToken(value,null);
    }
    
    public static boolean isToken(String value, String literals) {
        String tspecials =
                ((literals ==null) ? ServerCookie.tspecials : literals);
	if (value == null) {
            return true;
        }
	int len = value.length();

	for (int i = 0; i < len; i++) {
	    char c = value.charAt(i);

	    if (tspecials.indexOf(c) != -1) {
		return false;
            }
	}
	return true;
    }

    public static boolean containsCTL(String value, int version) {
        if (value == null) {
            return false;
        }
        int len = value.length();
        for (int i = 0; i < len; i++) {
            char c = value.charAt(i);
            if (c < 0x20 || c >= 0x7f) {
                if (c == 0x09)
                    continue; //allow horizontal tabs
                return true;
            }
        }
        return false;
    }

    public static boolean isToken2(String value) {
        return isToken2(value,null);
    }

    public static boolean isToken2(String value, String literals) {
        String tspecials2 =
                ((literals == null) ? ServerCookie.tspecials2 : literals);
        if (value == null) {
            return true;
        }
        int len = value.length();

        for (int i = 0; i < len; i++) {
            char c = value.charAt(i);

            if (tspecials2.indexOf(c) != -1)
                return false;
        }
        return true;
    }

    /**
     * @deprecated - Not used
     */
    public static boolean checkName( String name ) {
	if (!isToken(name)
                || name.equalsIgnoreCase("Comment")	// rfc2019
                || name.equalsIgnoreCase("Discard")	// rfc2965
                || name.equalsIgnoreCase("Domain")      // rfc2019
                || name.equalsIgnoreCase("Expires")	// Netscape
                || name.equalsIgnoreCase("Max-Age")	// rfc2019
                || name.equalsIgnoreCase("Path")        // rfc2019
                || name.equalsIgnoreCase("Secure")      // rfc2019
                || name.equalsIgnoreCase("Version")     // rfc2019
                // TODO remaining RFC2965 attributes
	    ) {
	    return false;
	}
	return true;
    }

    // -------------------- Cookie parsing tools

    
    /**
     * Return the header name to set the cookie, based on cookie version
     */
    public String getCookieHeaderName() {
	return getCookieHeaderName(version);
    }

    /** Return the header name to set the cookie, based on cookie
     *  version
     */
    public static String getCookieHeaderName(int version) {
        // TODO Re-enable logging when RFC2965 is implemented
        // log( (version==1) ? "Set-Cookie2" : "Set-Cookie");
        if (version == 1) {
            // XXX RFC2965 not referenced in Servlet Spec
            // Set-Cookie2 is not supported by Netscape 4, 6, IE 3, 5
            // Set-Cookie2 is supported by Lynx and Opera
            // Need to check on later IE and FF releases but for now...
            // RFC2109
            return "Set-Cookie";
            // return "Set-Cookie2";
        } else {
            // Old Netscape
            return "Set-Cookie";
        }
    }


    public static void appendCookieValue(StringBuffer buf,
                                         int version,
                                         String name,
                                         String value,
                                         String path,
                                         String domain,
                                         String comment,
                                         int maxAge,
                                         boolean isSecure) {
        appendCookieValue(buf, version, name, value, path, domain, comment,
                          maxAge, isSecure, false);
    }

    // TODO RFC2965 fields also need to be passed
    public static void appendCookieValue(StringBuffer headerBuf,
                                         int version,
                                         String name,
                                         String value,
                                         String path,
                                         String domain,
                                         String comment,
                                         int maxAge,
                                         boolean isSecure,
                                         boolean encode) {

        StringBuffer buf = new StringBuffer();
        // this part is the same for all cookies
        if (encode) {
            try {
                // Servlet implementation checks name
                buf.append(URLEncoder.encode(name, "UTF-8"));
                buf.append("=");
                // Servlet implementation does not check anything else
                /* PWC 6392327
                maybeQuote(version, buf, URLEncoder.encode(value, "UTF-8"));
                */
                // START PWC 6392327
                if (value != null) {
                    maybeQuote(version, buf,
                               URLEncoder.encode(value, "UTF-8"));
                } else {
                    maybeQuote(version, buf,
                               URLEncoder.encode(ServerCookie.NULL_VALUE,
                                                 "UTF-8"));
                }
                // END PWC 6392327
            } catch (UnsupportedEncodingException e) {
                buf.append(URLEncoder.encode(name));
                buf.append("=");
                /* PWC 6392327
                maybeQuote(version, buf, URLEncoder.encode(value));
                */
                // START PWC 6392327
                if (value != null) {
                    maybeQuote(version, buf,
                               URLEncoder.encode(value));
                } else {
                    maybeQuote(version, buf,
                               URLEncoder.encode(ServerCookie.NULL_VALUE));
                }
                // END PWC 6392327
            }
        } else {
            buf.append( name );
            buf.append("=");
            /* PWC 6392327
            maybeQuote(version, buf, value);
            */
            // START PWC 6392327
            if (value != null) {
                maybeQuote(version, buf, value);
            } else {
                maybeQuote(version, buf, ServerCookie.NULL_VALUE);
            }
            // END PWC 6392327
        }

 	// Add version 1 specific information
	if (version == 1) {
	    // Version=1 ... required
	    buf.append ("; Version=1");

	    // Comment=comment
	    if (comment!=null) {
                buf.append ("; Comment=");
                if (encode) {
                    try {
                        maybeQuote(version, buf,
                                   URLEncoder.encode(comment, "UTF-8"));
                    } catch(UnsupportedEncodingException e) {
                        maybeQuote(version, buf,
                                   URLEncoder.encode(comment));
                    }
	        } else {
                    maybeQuote (version, buf, comment);
                }
            }
        }
	
	// Add domain information, if present

	if (domain!=null) {
	    buf.append("; Domain=");
	    maybeQuote (version, buf, domain);
	}

	// Max-Age=secs ... or use old "Expires" format
        // TODO RFC2965 Discard
	if (maxAge >= 0) {
	    if (version == 0) {
                // Wdy, DD-Mon-YY HH:MM:SS GMT ( Expires Netscape format )
		buf.append ("; Expires=");
                // To expire immediately we need to set the time in past
                if (maxAge == 0)
		    buf.append( ANCIENT_DATE );
		else
                    DateTool.formatOldCookie
                        (new Date( System.currentTimeMillis() +
                                   maxAge *1000L), buf,
                         new FieldPosition(0));

	    } else {
		buf.append ("; Max-Age=");
		buf.append (maxAge);
	    }
	}

	// Path=path
	if (path!=null) {
	    buf.append ("; Path=");
            if (version == 0) {
                maybeQuote(version, buf, path);
            } else {
                maybeQuote(version, buf, path, ServerCookie.tspecials2NoSlash);
            }
	}

	// Secure
	if (isSecure) {
	  buf.append ("; Secure");
	}
	
        headerBuf.append(buf);
    }


    // START OF IASRI 4830338
    public static void appendEncodedCookieValue( StringBuffer buf,
					  int version,
					  String name,
					  String value,
					  String path,
					  String domain,
					  String comment,
					  int maxAge,
					  boolean isSecure) {
        appendCookieValue(buf, version, name, value, path, domain, comment,
                          maxAge, isSecure, true);
    }
    // END OF IASRI 4830338

    public static void maybeQuote(int version, StringBuffer buf, String value) {
        maybeQuote(version, buf, value, null);
    }

    /**
     * Quotes values using rules that vary depending on Cookie version.
     * @param version
     * @param buf
     * @param value
     * @param literals
     */
    public static void maybeQuote (int version, StringBuffer buf,
                                   String value, String literals)
    {
        if (value == null || value.length() == 0) {
            buf.append("\"\"");
        } else if (containsCTL(value,version)) { 
            throw new IllegalArgumentException(
                sm.getString("servercookie.control_character"));
        } else if (alreadyQuoted(value)) {
            buf.append('"');
            buf.append(escapeDoubleQuotes(value,1,value.length()-1));
            buf.append('"');
            buf.append('"');
        } else if (version==0 && !isToken(value,literals)) {
            buf.append('"');
            buf.append(escapeDoubleQuotes(value,0,value.length()));
            buf.append('"');
        } else if (version==1 && !isToken2(value,literals)) {
            buf.append('"');
            buf.append(escapeDoubleQuotes(value,0,value.length()));
            buf.append('"');
        } else {
            buf.append(value);
        }
    }

    // log
    static final int dbg=1;
    private static void log(String s ) {
        if (logger.isLoggable(Level.FINE))
            logger.fine("Cookies: " + s);
    }

    public static boolean alreadyQuoted (String value) {
        if (value == null || value.length() == 0) {
            return false;
        }
        return (value.charAt(0) == '\"' && value.charAt(value.length()-1) == '\"');
    }

    /**
     * Escapes any double quotes in the given string.
     *
     * @param s the input string
     * @param beginIndex start index inclusive
     * @param endIndex exclusive
     * @return The (possibly) escaped string
     */
    private static String escapeDoubleQuotes(String s, int beginIndex,
            int endIndex) {

        if (s == null || s.length() == 0 || s.indexOf('"') == -1) {
            return s;
        }

        StringBuffer b = new StringBuffer();
        for (int i = beginIndex; i < endIndex; i++) {
            char c = s.charAt(i);
            if (c == '\\' ) {
                b.append(c);
                //ignore the character after an escape, just append it
                if (++i >= endIndex) {
                    throw new IllegalArgumentException(
                        sm.getString("servercookie.invalid_escape"));
                }
                b.append(s.charAt(i));
            } else if (c == '"') {
                 b.append('\\').append('"');
            } else {
                 b.append(c);
            }
        }

        return b.toString();
    }

   /**
     * Unescapes any double quotes in the given cookie value.
     *
     * @param bc The cookie value to modify
     */
    public static void unescapeDoubleQuotes(ByteChunk bc) {

        if (bc == null || bc.getLength() == 0 || bc.indexOf('"', 0) == -1) {
            return;
        }

        int src = bc.getStart();
        int end = bc.getEnd();
        int dest = src;
        byte[] buffer = bc.getBuffer();
        
        while (src < end) {
            if (buffer[src] == '\\' && src < end && buffer[src+1]  == '"') {
                src++;
            }
            buffer[dest] = buffer[src];
            dest++;
            src++;
        }
        bc.setEnd(dest);
    }
}
